/* $Id: nit3open.c,v 1.2 89/07/11 03:07:44 dupuy Exp $ */

/*
 * Returns file descriptor for ethernet device by name ("ie0", "le0", etc.).
 * If name is NULL, uses primary ethernet interface.  Will only receive packets
 * of type specified.  Will receive packets for the ethernet address specified,
 * or local ethernet address if NULL.  If there is an error, returns (-1) and
 * the appropriate value is left in errno.  Normal return status zero.
 * Requires superuser privilege.
 */

#include <strings.h>
#include <errno.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/nit.h>
#include <netinet/in.h>			/* htons() macro */

#include "libether.h"

struct timeval *ether_timeout;

static ether_addr promiscuous;

static fd_set multicast;
static gotlocal;
static ether_addr local_addr;
static ether_addr bcast_addr = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
static ether_addr ether_multi_addrs[NOFILE];
static short ether_types[NOFILE];

#define multi_addr (ether_multi_addrs[fd])
#define ether_type (ether_types[fd])


int
ether_open (name, type, address)
char *name;
int type;
ether_addr *address;
{
    int fd;
    struct sockaddr_nit snit;
    struct nit_ioc nioc;
    char **interfaces;
    int saved_errno;

    if (name == 0)
    {					/* get default ethernet interface */
	interfaces = ether_interfaces ();
	if (interfaces == 0 || *interfaces == 0)
	    return (-1);

	name = *interfaces;		/* just use the first name in list */
    }

    if ((fd = socket (AF_NIT, SOCK_RAW, NITPROTO_RAW)) < 0)
    {
#ifdef DEBUG
	perror ("ether_open: socket");
#endif
	return (-1);
    }

    snit.snit_family = AF_NIT;
    strncpy (snit.snit_ifname, name, NITIFSIZ);

    if (bind (fd, (struct sockaddr *) &snit, sizeof (snit)) != 0)
    {
	saved_errno = errno;
#ifdef DEBUG
	perror ("ether_open: bind");
#endif
	close (fd);
	errno = saved_errno;
	return (-1);
    }

    bzero ((char *) &nioc, sizeof (nioc));

    nioc.nioc_bufalign = sizeof (long);
    nioc.nioc_chunksize = nioc.nioc_bufspace = ETHER_BUFSIZ;
    nioc.nioc_typetomatch = htons (type);
    nioc.nioc_snaplen = ETHER_PKT + ETHER_MAX;

    if (address == 0)			/* not a multicast address */
    {
	multicast.fds_bits[0] &= ~(1 << fd);
	nioc.nioc_flags = NF_TIMEOUT;
    }
    else				/* a multicast address */
    {
	if (bcmp ((char *) address, (char *) &promiscuous, sizeof (*address)))
	{
	    if (!ETHER_MCAST (address))
	    {
#ifdef DEBUG
		printf ("rejecting non-multicast address argument\n");
#endif
		close (fd);
		errno = EINVAL;
		return (-1);
	    }

	    multicast.fds_bits[0] |= (1 << fd);
	    multi_addr = *address;
	    if (gotlocal == 0)
		if (ether_address (fd, &local_addr) != NULL)
		    gotlocal = 1;
	}
	else
	    multicast.fds_bits[0] &= ~(1 << fd);

	nioc.nioc_flags = NF_TIMEOUT | NF_PROMISC;
    }

    if (ether_timeout != 0)		/* use specified value of timeout */
	nioc.nioc_timeout = *ether_timeout;
    else
    {					/* then use default value of timeout */
	nioc.nioc_timeout.tv_sec = 0;
	nioc.nioc_timeout.tv_usec = 10000;
    }

    if (ioctl (fd, SIOCSNIT, &nioc) != 0)
    {
	saved_errno = errno;
#ifdef DEBUG
	perror ("ether_open: ioctl SIOCSNIT");
#endif
	close (fd);
	errno = saved_errno;
	return (-1);
    }

    ether_type = type;
    return (fd);
}


#ifdef DEBUG

void
dump_nit_hdr (nitbuf)
struct nit_hdr *nitbuf;
{
    switch (nitbuf->nh_state)
    {
      case NIT_QUIET:
	printf ("ether_read: nit quiet\n");
	break;

      case NIT_CATCH:
	printf ("ether_read: nit catch: size %d\n", nitbuf->nh_datalen);
	break;

      case NIT_NOMBUF:
	printf ("ether_read: nit no mbufs: %d lost\n", nitbuf->nh_dropped);
	break;

      case NIT_NOCLUSTER:
	printf ("ether_read: nit no mclusters: %d lost\n", nitbuf->nh_dropped);
	break;

      case NIT_NOSPACE:
	printf ("ether_read: nit no bufspace: %d lost\n", nitbuf->nh_dropped);
	break;

      case NIT_SEQNO:
	printf ("ether_read: nit sequence # %d\n", nitbuf->nh_seqno);
	break;

      default:
	printf ("ether_read: bad nit header state: %d\n", nitbuf->nh_state);
	break;
    }
}

#endif


/*
 * Reads and returns a single packet, filling in all fields.  If pktbuf is
 * NULL, a buffer is allocated for it.	If pktbuf is not NULL, the function
 * assumes that pktbuf is large enough to hold the packet.  Since read() may
 * return several packets at a time, ether_read() has to buffer them as well.
 * Since socket based nit doesn't handle multicast addresses, we have to check
 * them at the user level.
 */

static long ether_buf[NOFILE][ETHER_BUFSIZ / sizeof (long)];
static int ether_buflen[NOFILE];
static char *ether_bufptr[NOFILE];

#define buf	(ether_buf[fd])
#define buflen	(ether_buflen[fd])
#define bufptr	(ether_bufptr[fd])

int
ether_read (fd, packet)
int fd;
ether_packet *packet;
{
    register struct nit_hdr *nitbuf;
    register char *pbuf;
    register int count;

    if (bufptr == 0)			/* easier than static initialization */
	bufptr = (char *) buf;

    for (;;)
    {
	/* begin critical section */

	if ((bufptr - (char *) buf) >= buflen)
	{				/* then buffer is used up */
	    buflen = read (fd, (char *) buf, ETHER_BUFSIZ);
	    bufptr = (char *) buf;
	    if (buflen <= 0)
	    {
		if (errno = EWOULDBLOCK)
		    errno = EAGAIN;
		return (-1);
	    }
	}

	nitbuf = (struct nit_hdr *) bufptr;
	while (nitbuf->nh_state != NIT_CATCH)
	{
#ifdef DEBUG
	    dump_nit_hdr (nitbuf);
#endif
	    nitbuf++;
	    if ((char *) nitbuf - (char *) buf > buflen)
	    {
#ifdef DEBUG
		printf ("truncated nit_hdr struct in buffer\n");
#endif
		continue;
	    }
	}
#ifdef DEBUG
	dump_nit_hdr (nitbuf);
#endif
	bufptr = (pbuf = (char *) (nitbuf + 1)) + nitbuf->nh_datalen +
	    pad (nitbuf->nh_datalen);

	/* end critical section */

	if ((multicast.fds_bits[0] & (1 << fd)) &&
	    ((! ETHER_MCAST (pbuf) 
	      && bcmp (pbuf, (char *) &local_addr, sizeof (ether_addr)) != 0) 
	     ||
	    (bcmp (pbuf, (char *) &bcast_addr, sizeof (ether_addr)) != 0
	     && bcmp (pbuf, (char *) &multi_addr, sizeof (ether_addr)) != 0)))
	{
#ifdef DEBUG
	    printf ("discarding incorrectly addressed packet\n");
#endif
	    continue;
	}

	if (nitbuf->nh_wirelen > nitbuf->nh_datalen)
	{
#ifdef DEBUG
	    printf ("discarding truncated packet\n");
#endif
	    continue;
	}
	break;
    }

    bcopy (pbuf, (char *) packet, ETHER_PKT);
    pbuf += ETHER_PKT;

    if ((count = nitbuf->nh_wirelen - ETHER_PKT) < 0)
	errno = EBADMSG;
    else
    {
	if (packet->pktbuf == NULL)
	{
	    packet->pktlen = count;
	    if ((packet->pktbuf = malloc ((unsigned) count)) == NULL)
		return (-1);
	}

	bcopy (pbuf, packet->pktbuf,
	       (int) ((packet->pktlen > count) ? count : packet->pktlen));

	if (packet->pktlen > count)
	    packet->pktlen = count;
    }

    return (count);
}


/*
 * Writes a single packet, filling src and type fields.
 */

int 
ether_write (fd, packet)
int fd;
ether_packet *packet;
{
    struct sockaddr sa;

    sa.sa_family = AF_UNSPEC;
    bcopy ((char *) packet, sa.sa_data, ETHER_TYPE);

    sa.sa_data[12] = (ether_type >> 8) & 0xff;
    sa.sa_data[13] = ether_type & 0xff;

    return (sendto (fd, packet->pktbuf, (int) packet->pktlen, 0, &sa,
		    sizeof (sa)) < 0 ? -1 : 0);
}
